@@ -3,10 +3,11 @@ current_dir = os.path.dirname(os.path.realpath(__file__)) + "/"
 
 filenames = [
     "ex-benchmark-and-fork/ex-benchmark-and-fork.py", "ex-fodo/ex-fodos.py", "ex-lhc-couplingLocal/lhc-couplingLocal.py",
-    "ex-limitations/ex-limitations.py", "ex-LowLevel/ex-send-multypes.py", "ex-LowLevel/ex-send-recv.py", 
+    "ex-managing-refs/ex-managing-refs.py", "ex-LowLevel/ex-send-multypes.py", "ex-LowLevel/ex-send-recv.py", 
     "ex-ps-twiss/ps-twiss.py", "ex-recv-lhc/ex-defexpr.py"
     ]
 for name in filenames:
+    print(name)
     env = {"__file__": current_dir + name}
     with open(name, "r") as f:
         exec(f.read(), env)
