@@ -1,9 +1,7 @@
 from .mad_object import MAD
 
 __title__ = "pymadng"
-__version__ = "0.3.3"
-with MAD() as mad:
-    __MAD_version__ = mad.MAD.env.version
+__version__ = "0.3.5"
 
 __summary__ = "Python interface to MAD-NG running as subprocess"
 __uri__ = "https://github.com/MethodicalAcceleratorDesign/MADpy"
